@@ -20,6 +20,7 @@ unset LOCAL             || true
 unset LOCALDEBS         || true
 unset SECURITY          || true
 unset PROPOSED_UPDATES  || true
+unset UNRELEASED        || true
 unset BOOTDIR           || true
 unset SYMLINK           || true
 unset COPYLINK          || true
@@ -61,13 +62,23 @@ export BASEDIR=`pwd`
 # Name of the distribution for the image filename (Defaults to 'debian')
 # export CDNAME=debian
 
-# Building wheezy cd set ...
-export CODENAME=wheezy
+# Building $codename cd set ...
+export CODENAME=buster
 
 # By default use Debian installer packages from $CODENAME
 if [ -z "$DI_CODENAME" ]; then
 	export DI_CODENAME=$CODENAME
 fi
+# If you want backported d-i (e.g. by setting
+# DI_CODENAME=jessie-backports, then you'll almost definitely also
+# want to enable BACKPORTS below as well
+
+# Should we include some packages from backports? If so, point at a
+# file containing a list of binary packages to include here. Packages
+# contained in that list will be chosen by preference from
+# ${CODENAME}-backports rather than just ${CODENAME}, along with
+# dependencies if they're needed as well
+#export BACKPORTS=backports-list
 
 # If set, controls where the d-i components are downloaded from.
 # This may be an url, or "default", which will make it use the default url
@@ -76,7 +87,7 @@ fi
 #export DI_WWW_HOME=default
 
 # Version number, "2.2 r0", "2.2 r1" etc.
-export DEBVERSION="7.0"
+export DEBVERSION="10.0.0"
 
 # Official or non-official set.
 # NOTE: THE "OFFICIAL" DESIGNATION IS ONLY ALLOWED FOR IMAGES AVAILABLE
@@ -156,6 +167,9 @@ export CONTRIB=1
 # NOTE: intended to be used for pre-release testing, not for publication!
 #export PROPOSED_UPDATES=$CODENAME-proposed-updates
 
+# Include packages from Debian Ports unreleased
+#export UNRELEASED=1
+
 # Sparc only : bootdir (location of cd.b and second.b)
 # export BOOTDIR=/boot
 
@@ -410,6 +424,13 @@ case "$OFFICIAL" in
 	;;
 esac
 
+# Add options to wget to include support for the Debian CA, so
+# https://d-i.debian.org et al will work.
+if [ -d "/etc/ssl/ca-debian" ]; then
+	export WGET_OPTS="--ca-directory /etc/ssl/ca-debian/"
+fi
+export WGET="wget $WGET_OPTS"
+
 # Set this to force the Release file(s) to say "stable". Used in first
 # Etch builds to allow us to build before the archive updated
 #EARLY_BUILD_HACK=1
