@@ -9,6 +9,7 @@ use Digest::MD5;
 use Digest::SHA;
 use File::stat;
 use File::Find;
+use File::Path qw(make_path remove_tree);
 use File::Basename;
 use Compress::Zlib;
 
@@ -47,6 +48,11 @@ my $localdebs = read_env('LOCALDEBS', $mirror);
 my $symlink_farm = read_env('SYMLINK', 0);
 my $link_verbose = read_env('VERBOSE', 0);
 my $link_copy = read_env('COPYLINK', 0);
+my $backports_list = read_env('BACKPORTS', "");
+my $backports = 1;
+if ($backports_list =~ /^$/) {
+    $backports = 0;;
+}
 
 # MAXCDS is the hard limit on the MAXIMUM number of images to
 # make. MAXJIGDOS and MAXISOS can only make this number smaller; we
@@ -100,17 +106,9 @@ foreach my $arch (split(' ', $archlist)) {
     load_packages_cache($arch);
 }
 
-if (! ($archlist eq "source")) {
-    load_descriptions("main");
-    if ($contrib) {
-        load_descriptions("contrib");
-    }
-    if ($nonfree || $extranonfree) {
-        load_descriptions("non-free");
-    }
-    if ($use_local) {
-        load_descriptions("local");
-    }
+load_all_descriptions(0);
+if ($backports) {
+    load_all_descriptions(1);
 }
 
 my $disknum = 1;
@@ -342,6 +340,7 @@ close(LOG);
 sub load_packages_cache {
     my $arch = shift;
     my @pkglist;
+    my @tmplist;
     my ($p);
     my $num_pkgs = 0;
 
@@ -354,7 +353,7 @@ sub load_packages_cache {
     while (defined (my $pkg = <INLIST>)) {
         chomp $pkg;
         my ($junk, $component, $pkgname, $pkgsize) = split /:/, $pkg;
-        push @pkglist, $pkgname;
+        push @tmplist, $pkgname;
     }
     close INLIST;
 
@@ -362,6 +361,7 @@ sub load_packages_cache {
     print LOG "Reading in package information for $arch:\n";
 
     $/ = ''; # Browse by paragraph
+    @pkglist = (grep (!/\/$codename-backports$/, @tmplist));
     while (@pkglist) {
         my (@pkg) = splice(@pkglist,0,200);
         if ($arch eq "source") {
@@ -379,24 +379,73 @@ sub load_packages_cache {
         close LIST;
         print LOG "load_packages_cache: Read details of $num_pkgs packages for $arch\n";
     }
-    $/ = $old_split; # Browse by line again
     print "  Done: Read details of $num_pkgs packages for $arch\n";
+    if ($backports) {
+	$num_pkgs = 0;
+	@pkglist = (grep (/\/$codename-backports$/, @tmplist));
+	while (@pkglist) {
+	    my (@pkg) = splice(@pkglist,0,200);
+	    if ($arch eq "source") {
+		open (LIST, "USE_BP=1 $basedir/tools/apt-selection cache showsrc @pkg |")
+		    || die "Can't fork : $!\n";
+	    } else {
+		open (LIST, "USE_BP=1 $basedir/tools/apt-selection cache show @pkg |")
+		    || die "Can't fork : $!\n";
+	    }
+	    while (defined($_ = <LIST>)) {
+		m/^Package: (\S+)/m and $p = $1;
+		push @{$pkginfo{$arch}{"$p/$codename-backports"}}, $_;
+		$num_pkgs++;
+	    }
+	    close LIST;
+	    print LOG "load_packages_cache: Read details of $num_pkgs packages for $arch backports\n";
+	}
+	print "  Done: Read details of $num_pkgs packages for $arch backports\n";
+    }
+    $/ = $old_split; # Browse by line again
+}
+
+sub load_all_descriptions {
+    my $use_backports = shift;
+    if (! ($archlist eq "source")) {
+	load_descriptions("main", $use_backports);
+	if ($contrib) {
+	    load_descriptions("contrib", $use_backports);
+	}
+	if ($nonfree || $extranonfree) {
+	    load_descriptions("non-free", $use_backports);
+	}
+	if ($use_local) {
+	    load_descriptions("local", $use_backports);
+	}
+    }
 }
 
 # Load all the translated descriptions we can find
 sub load_descriptions {
     my $suite = shift;
+    my $use_backports = shift;
     my $lang;
 	my $dh;
     my ($p);
     my $num_total = 0;
     my $num_files = 0;
     my $dir = "$mirror/dists/$codename/$suite/i18n";
+    if ($use_backports) {
+	$dir = "$mirror/dists/$codename-backports/$suite/i18n";
+    }
     if ($suite =~ /local/) {
         $dir = "$localdebs/dists/$codename/$suite/i18n";
+	if ($use_backports) {
+	    $dir = "$localdebs/dists/$codename-backports/$suite/i18n";
+	}
     }
     my @files;
 
+    if ($use_backports) {
+	$suite = "$suite backports";
+    }
+
     if (-d $dir) {
         print "Reading in translated package descriptions for $suite:\n";
         print LOG "Reading in translated package descriptions for $suite:\n";
@@ -413,6 +462,9 @@ sub load_descriptions {
                 print LOG "  Parsing $file\n";
                 while (defined($_ = <BZ>)) {
                     m/^Package: (\S+)/m and $p = $1;
+		    if ($use_backports) {
+			$p = "$p/$codename-backports";
+		    }
                     $descriptions{"$lang"}{$p}{"data"} = $_;
                     $descriptions{"$lang"}{$p}{"used"} = 0;
                     $num_descs++;
@@ -534,7 +586,7 @@ sub check_base_installable {
         print LOG "Debootstrap reported error: $error_string\n";
         die "Debootstrap reported error: $error_string\n";
     }
-	system("rm -rf $tdir/debootstrap_tmp");
+	remove_tree("$tdir/debootstrap_tmp");
 	return $ok;
 }
 
@@ -814,7 +866,8 @@ sub finish_disc {
 			$ok += $archok;
 		}
 		if ($ok == 0) {
-			system("touch $cddir/.disk/base_installable");
+			open(my $fh, ">>", "$cddir/.disk/base_installable");
+			close($fh);
 			print "  Found all files needed for debootstrap for all binary arches\n";
 		} else {
 			print "  $ok files missing for debootstrap, not creating base_installable\n";
@@ -836,13 +889,19 @@ sub finish_disc {
 	}
 
 	print "  Finishing off the Release file\n";
-	chdir "dists/$codename";
-	open(RELEASE, ">>Release") or die "Failed to open Release file: $!\n";
-	find (\&recompress, ".");
-	checksum_files_for_release();
-	close(RELEASE);
-	find (\&remove_uncompressed, ".");
-	chdir("../..");
+	my @codenames = ("$codename");
+	if ($backports) {
+	    push @codenames, "$codename-backports";
+	}
+	foreach my $tmpcode (@codenames) {
+	    chdir "dists/$tmpcode";
+	    open(RELEASE, ">>Release") or die "Failed to open Release file: $!\n";
+	    find (\&recompress, ".");
+	    checksum_files_for_release();
+	    close(RELEASE);
+	    find (\&remove_uncompressed, ".");
+	    chdir("../..");
+	}
 
 	print "  Finishing off md5sum.txt\n";
 	# Just md5 the bits we won't have seen already
@@ -853,7 +912,7 @@ sub finish_disc {
 	# And sort; it should make things faster for people checking
 	# the md5sums, as ISO9660 dirs are sorted alphabetically
 	system("LANG=C sort -uk2 md5sum.txt | grep -v \./md5sum.txt > md5sum.txt.tmp");
-	system("mv -f md5sum.txt.tmp md5sum.txt");
+	rename("md5sum.txt.tmp", "md5sum.txt");
 	chdir $bdir;
 
 	if (defined($ENV{'DISC_END_HOOK'})) {
@@ -878,6 +937,7 @@ sub msg_ap {
     if (!$log_opened) {
         open(AP_LOG, ">> $tdir/$codename/add_packages.log")
             || die "Can't write in $tdir/add_packages.log!\n";
+	$log_opened = 1;
     }
     print AP_LOG @_;
 }
@@ -893,6 +953,7 @@ sub Packages_dir {
     my $dir = shift;
     my $file = shift;
     my $section = shift;
+    my $in_backports = shift;
 
     my ($pdir, $dist);
 
@@ -907,8 +968,13 @@ sub Packages_dir {
     }	
 
     $pdir = "$dir/dists/$codename/$dist";
+    if ($in_backports) {
+	$pdir = "$dir/dists/$codename-backports/$dist";
+    }	
     if ($section and $section eq "debian-installer") {
         $pdir = "$dir/dists/$codename/$dist/debian-installer";
+	# Don't attempt to put d-i components into backports, as d-i
+	# won't look for them there.
     }
     return $pdir;
 }
@@ -918,7 +984,8 @@ sub Packages_dir {
 sub add_Packages_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my ($p, $file, $section, $pdir, $pkgfile, $gz, $st1, $st2, $size1, $size2);
     my $blocks_added = 0;
     my $old_blocks = 0;
@@ -932,21 +999,21 @@ sub add_Packages_entry {
         if (!defined($file)) {
             die "Can't parse source file information out of $_\n";
         }
-        $pdir = Packages_dir($dir, $file, $section) . "/source";
+        $pdir = Packages_dir($dir, $file, $section, $in_backports) . "/source";
         $pkgfile = "$pdir/Sources";
     } else {
         m/^Filename: (\S+)/mi and $file = $1;
         if (!defined($file)) {
             die "Can't parse binary file information out of $_\n";
         }
-        $pdir = Packages_dir($dir, $file, $section) . "/binary-$arch";
+        $pdir = Packages_dir($dir, $file, $section, $in_backports) . "/binary-$arch";
         $pkgfile = "$pdir/Packages";
     }
 
     msg_ap(0, "  Adding $p to $pkgfile(.gz)\n");
     
     if (! -d $pdir) {
-        system("mkdir -p $pdir");
+        make_path($pdir);
         $blocks_added++;
     }	
 
@@ -983,7 +1050,8 @@ sub add_Packages_entry {
 sub add_trans_desc_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my ($p, $file, $section, $idir, $pkgfile, $gz, $st);
     my $size = 0;
     my $blocks_added = 0;
@@ -994,10 +1062,10 @@ sub add_trans_desc_entry {
     m/^Section: (\S+)/m and $section = $1;
 
     m/^Filename: (\S+)/mi and $file = $1;
-    $idir = Packages_dir($dir, $file, $section) . "/i18n";
+    $idir = Packages_dir($dir, $file, $section, $in_backports) . "/i18n";
 
     if (! -d $idir) {
-        system("mkdir -p $idir");
+        make_path($idir);
         $blocks_added++;
     }	
 
@@ -1042,21 +1110,28 @@ sub add_trans_desc_entry {
         }
     }
     $blocks_added += ($new_blocks - $old_blocks);
-    msg_ap(0, "    now $size bytes, $blocks_added blocks added\n");
+    if ($blocks_added != 0) {
+	msg_ap(0, "    now $size bytes, $blocks_added blocks added\n");
+    }
     return $blocks_added;
 }
 
 sub add_md5_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my ($pdir, $file, $md5, $st, $size, $p);
     my $md5file = "$dir/md5sum.txt";
     my $blocks_added = 0;
     my $old_blocks = 0;
     my $new_blocks = 0;
 
-    m/^Package: (\S+)/mi and $p = $1;
+    if ($in_backports) {
+	m/^Package: (\S+)/mi and $p = "$1/$codename-backports";
+    } else {
+	m/^Package: (\S+)/mi and $p = $1;
+    }
 
     if (-e $md5file) {
         $st = stat("$md5file");
@@ -1089,7 +1164,8 @@ sub add_md5_entry {
 sub remove_Packages_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my ($p, $file, $section, $pdir, $pkgfile, $tmp_pkgfile, $match, $gz,
         $st1, $st2, $size1, $size2);
     my $blocks_removed = 0;
@@ -1101,11 +1177,11 @@ sub remove_Packages_entry {
 
     if ($arch eq "source") {
         m/^Directory: (\S+)/mi and $file = $1;
-        $pdir = Packages_dir($dir, $file, $section) . "/source";
+        $pdir = Packages_dir($dir, $file, $section, $in_backports) . "/source";
         $pkgfile = "$pdir/Sources";
     } else {
         m/^Filename: (\S+)/mi and $file = $1;
-        $pdir = Packages_dir($dir, $file, $section) . "/binary-$arch";
+        $pdir = Packages_dir($dir, $file, $section, $in_backports) . "/binary-$arch";
         $pkgfile = "$pdir/Packages";
     }
 
@@ -1157,7 +1233,8 @@ sub remove_Packages_entry {
 sub remove_trans_desc_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my ($p, $file, $section, $idir, $gz, $match, $st);
     my $size = 0;
     my $blocks_added = 0;
@@ -1168,7 +1245,7 @@ sub remove_trans_desc_entry {
     m/^Section: (\S+)/m and $section = $1;
 
     m/^Filename: (\S+)/mi and $file = $1;
-    $idir = Packages_dir($dir, $file, $section) . "/i18n";
+    $idir = Packages_dir($dir, $file, $section, $in_backports) . "/i18n";
 
     $/ = ''; # Browse by paragraph
     foreach my $lang (keys %descriptions) {
@@ -1226,7 +1303,8 @@ sub remove_trans_desc_entry {
 sub remove_md5_entry {
     my $dir = shift;
     my $arch = shift;
-    my $_ = shift;
+    my $in_backports = shift;
+    local $_ = shift;
     my $md5file = "$dir/md5sum.txt";
     my $tmp_md5file = "$dir/md5sum.txt.tmp";
     my @fileslist;
@@ -1315,6 +1393,10 @@ sub add_packages {
     msg_ap(0, "Looking at $pkg: arch $arch, package $pkgname, rollback $rollback\n");
 
     foreach my $package_info (@{$pkginfo{$arch}{$pkgname}}) {
+	my $in_backports = 0;
+	if ($pkgname =~ /\/$codename-backports/) {
+	    $in_backports = 1;
+	}
         undef @files;
         $source = $mirror;
         if ($arch eq "source") {
@@ -1339,10 +1421,10 @@ sub add_packages {
 
         if ($rollback) {
             # Remove the Packages entry/entries for the specified package
-            $total_blocks -= remove_Packages_entry($dir, $arch, $package_info);
-            $total_blocks -= remove_md5_entry($dir, $arch, $package_info);
+            $total_blocks -= remove_Packages_entry($dir, $arch, $in_backports, $package_info);
+            $total_blocks -= remove_md5_entry($dir, $arch, $in_backports, $package_info);
             if (!($arch eq "source")) {
-                $total_blocks -= remove_trans_desc_entry($dir, $arch, $package_info);
+                $total_blocks -= remove_trans_desc_entry($dir, $arch, $in_backports, $package_info);
             }
         
             foreach my $file (@files) {
@@ -1364,10 +1446,10 @@ sub add_packages {
                 msg_ap(0, "  Rollback: removed $dir/$file\n");
             }
         } else {
-            $total_blocks += add_Packages_entry($dir, $arch, $package_info);
-            $total_blocks += add_md5_entry($dir, $arch, $package_info);
+            $total_blocks += add_Packages_entry($dir, $arch, $in_backports, $package_info);
+            $total_blocks += add_md5_entry($dir, $arch, $in_backports, $package_info);
             if (!($arch eq "source")) {
-                $total_blocks += add_trans_desc_entry($dir, $arch, $package_info);
+                $total_blocks += add_trans_desc_entry($dir, $arch, $in_backports, $package_info);
             }
 
             foreach my $file (@files) {
